local GameTick = 0

local TimeStopped = false
local FlashTick = 0

local Damaged = {}

local function Damage(target, inflictor)
	target.tsdmg = (target.tsdmg or 0) + 1
	target.tslastdmg = GameTick + 5
	target.tsinflictor = inflictor
	if not TS_InTable(Damaged, target) then
		table.insert(Damaged, target)
	end
end

local function ApplyDamage()
	for i,mo in pairs(Damaged) do
		if mo.valid then
			if mo.tsdmg then
				P_DamageMobj(mo, mo.tsinflictor, mo.tsinflictor, mo.tsdmg)
			end
			mo.tsdmg = nil
			mo.tslastdmg = nil
			mo.tsinflictor = nil
		end
	end
	Damaged = {}
end

COM_AddCommand("timestop", function(player, argument)
	if player.timestop then
		player.timestop.enabled = not player.timestop.enabled
		local message = (player.timestop.enabled and "Time Stop enabled!") or "Time Stop disabled!"
		CONS_Printf(player, "\x84" .. message)
	end
end, 0)

addHook("PlayerThink", function(player)
	if not player.timestop then
		player.timestop = {}
		player.timestop.enabled = true
		player.timestop.input = 0
	end
	
	if TimeStopped then
		if FlashTick > 0 then
			FlashTick = FlashTick - 1
			P_FlashPal(player, 2, 8)
		else
			P_FlashPal(player, 5, 8)
		end
		player.powers[pw_spacetime] = 0
		player.powers[pw_underwater] = 0
		if player.freezepowers == nil then
			player.freezepowers = {}
		end
		if player.powers[pw_invulnerability] and type(player.powers[pw_invulnerability]) == "number" then
			player.freezepowers.inv = player.freezepowers.inv or player.powers[pw_invulnerability]
			player.powers[pw_invulnerability] = player.freezepowers.inv
		end
		if player.powers[pw_sneakers] and type(player.powers[pw_sneakers]) == "number" then
			player.freezepowers.sneaker = player.freezepowers.sneaker or player.powers[pw_sneakers]
			player.powers[pw_sneakers] = player.freezepowers.sneaker
		end
	else
		if player.freezepowers then
			player.freezepowers = nil
		end
	end
	
	if not (player.mo and player.mo.valid) then
		return false
	end
	
	local timestop = player.timestop
	if timestop and timestop.enabled then
		if (player.cmd.buttons & BT_CUSTOM3) and player.timestop.input == 0 and not P_IsObjectOnGround(player.mo) and player.mo.state ~= S_PLAY_TIMESTOP then
			player.timestop.input = 1
			TimeStopped = not TimeStopped
			S_StartSound(nil, (TimeStopped and sfx_tsps) or sfx_tsrs)
			P_StartQuake(25*FRACUNIT, 15, {player.mo.x, player.mo.y, player.mo.z})
			if TimeStopped then
				S_StartSound(nil, sfx_tsfx)
				FlashTick = 5
				player.mo.state = S_PLAY_TIMESTOP
			else
				ApplyDamage()
				S_StartSound(nil, sfx_tsfz)
			end
		elseif not (player.cmd.buttons & BT_CUSTOM3) then
			player.timestop.input = 0
		end
	end
	
	if player.speed > 1*FRACUNIT and TimeStopped then
		local trail = P_SpawnGhostMobj(player.mo)
		trail.fuse = 35
	end
	if player.mo.state == S_PLAY_TIMESTOP then
		player.mo.momx = 0
		player.mo.momy = 0
		player.mo.momz = 0
	end
end)

addHook("ShouldDamage", function(target, inflictor, source, damage, damagetype)
	if inflictor and inflictor.valid and target.valid and (inflictor.flags & (MF_ENEMY|MF_BOSS)) and target.type == MT_PLAYER and TimeStopped then
		return false
	end
	if target.flags & (MF_ENEMY|MF_BOSS) and TimeStopped then
		if inflictor and inflictor.valid and inflictor.type == MT_PLAYER then
			if inflictor.player and inflictor.player.homing > 0 then
				inflictor.player.homing = 0
			end
		end
		return false
	end
end)

local function Collide(target,inflictor)
	if TimeStopped and inflictor.valid and target.valid and (target.flags & (MF_ENEMY|MF_BOSS)) and inflictor.type == MT_PLAYER and (inflictor.state == S_PLAY_JUMP or inflictor.state == S_PLAY_FALL or inflictor.state == S_PLAY_ROLL) then
		if target.tslastdmg == nil or GameTick > target.tslastdmg then
			Damage(target, inflictor)
			S_StartSound(target, sfx_dmpain)
		end
		if inflictor.player and inflictor.player.homing > 0 then
			inflictor.player.homing = 0
		end
		return
	end
end

addHook("MobjCollide", Collide)
addHook("MobjMoveCollide", Collide)

addHook("MapChange", function(mo)
	FlashTick = 0
	TimeStopped = false
end)

addHook("ThinkFrame", function()
	GameTick = GameTick + 1
end)

local function Thinker(mo)
	if mo.valid then
		if TimeStopped then
			return true
		end
	end
end

addHook("MobjThinker", function(mo)
	if mo.valid then
		if mo.flags & (MF_ENEMY|MF_BOSS|MF_MISSILE) and TimeStopped then
			return true
		end
	end
end)

for i,v in pairs(TS_Inclusion) do
	TS_NewHook(v, Thinker)
end

for i,v in pairs(TS_Flickies) do
	TS_NewHook(v, Thinker)
end